// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::CpuConfig,
    instruction::{AluBranchMOp, UnitOutRegNum},
    unit::{DynUnit, DynUnitWrapper, UnitCancelInput, UnitKind, UnitMOp, UnitOutput, UnitTrait},
};
use fayalite::{
    intern::{Intern, Interned},
    prelude::*,
    util::ready_valid::ReadyValid,
};

#[hdl_module]
pub fn alu_branch(config: &CpuConfig) {
    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let input: ReadyValid<AluBranchMOp<UnitOutRegNum<DynSize>, DynSize>> =
        m.input(ReadyValid[AluBranchMOp[config.unit_out_reg_num()][config.p_reg_num_width()]]);
    // TODO: finish
    connect(input.ready, true);
}

#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
pub struct AluBranch {
    config: Interned<CpuConfig>,
    module: Interned<Module<alu_branch>>,
}

impl AluBranch {
    pub fn new(config: &CpuConfig) -> Self {
        Self {
            config: config.intern(),
            module: alu_branch(config),
        }
    }
}

impl UnitTrait for AluBranch {
    type Type = alu_branch;
    type ExtraOut = ();
    type MOp = AluBranchMOp<UnitOutRegNum<DynSize>, DynSize>;

    fn ty(&self) -> Self::Type {
        self.module.io_ty()
    }

    fn extra_out_ty(&self) -> Self::ExtraOut {
        ()
    }

    fn mop_ty(&self) -> Self::MOp {
        self.module.io_ty().input.data.HdlSome
    }

    fn unit_kind(&self) -> UnitKind {
        UnitKind::AluBranch
    }

    fn extract_mop(
        &self,
        mop: Expr<UnitMOp<UnitOutRegNum<DynSize>, DynSize>>,
    ) -> Expr<HdlOption<Self::MOp>> {
        UnitMOp::alu_branch_mop(mop)
    }

    fn make_module(&self) -> Interned<Module<Self::Type>> {
        self.module
    }

    fn input(&self, this: Expr<Self::Type>) -> Expr<ReadyValid<Self::MOp>> {
        this.input
    }

    fn cancel_input(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<ReadyValid<UnitCancelInput<DynSize, DynSize>>> {
        todo!()
    }

    fn output(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<ReadyValid<UnitOutput<DynSize, DynSize, Self::ExtraOut>>> {
        todo!()
    }

    fn cd(&self, this: Expr<Self::Type>) -> Expr<ClockDomain> {
        this.cd
    }

    fn to_dyn(&self) -> DynUnit {
        DynUnitWrapper(*self).to_dyn()
    }
}
